@@ -23,10 +23,8 @@
 # THE SOFTWARE.
 #
 ###############################################################################
+from typing import Union
 
-from __future__ import absolute_import
-
-import six
 import txaio
 import inspect
 from functools import reduce
@@ -38,8 +36,8 @@ from autobahn.wamp import message
 from autobahn.wamp import types
 from autobahn.wamp import role
 from autobahn.wamp import exception
-from autobahn.wamp.exception import ApplicationError, ProtocolError, SessionNotReady, SerializationError
-from autobahn.wamp.interfaces import ISession, IPayloadCodec, IAuthenticator  # noqa
+from autobahn.wamp.exception import ApplicationError, ProtocolError, SessionNotReady, SerializationError, TypeCheckError
+from autobahn.wamp.interfaces import IPayloadCodec, IAuthenticator  # noqa
 from autobahn.wamp.types import SessionDetails, CloseDetails, EncodedPayload
 from autobahn.exception import PayloadExceededError
 from autobahn.wamp.request import \
@@ -147,8 +145,8 @@ class BaseSession(ObservableMixin):
             self._uri_to_ecls[exception._wampuris[0].uri()] = exception
         else:
             assert(not hasattr(exception, '_wampuris'))
-            self._ecls_to_uri_pat[exception] = [uri.Pattern(six.u(error), uri.Pattern.URI_TARGET_HANDLER)]
-            self._uri_to_ecls[six.u(error)] = exception
+            self._ecls_to_uri_pat[exception] = [uri.Pattern(error, uri.Pattern.URI_TARGET_HANDLER)]
+            self._uri_to_ecls[error] = exception
 
     def _message_from_exception(self, request_type, request, exc, tb=None, enc_algo=None):
         """
@@ -174,25 +172,19 @@ class BaseSession(ObservableMixin):
         if hasattr(exc, 'kwargs'):
             kwargs = exc.kwargs
 
-        if kwargs and six.PY2:
-            kwargs = {
-                k.decode('utf8'): v
-                for k, v in kwargs.iteritems()
-            }
-
         if tb:
             if kwargs:
-                kwargs[u'traceback'] = tb
+                kwargs['traceback'] = tb
             else:
-                kwargs = {u'traceback': tb}
+                kwargs = {'traceback': tb}
 
         if isinstance(exc, exception.ApplicationError):
-            error = exc.error if type(exc.error) == six.text_type else six.u(exc.error)
+            error = exc.error if type(exc.error) == str else exc.error
         else:
             if exc.__class__ in self._ecls_to_uri_pat:
                 error = self._ecls_to_uri_pat[exc.__class__][0]._uri
             else:
-                error = u"wamp.error.runtime_error"
+                error = "wamp.error.runtime_error"
 
         encoded_payload = None
         if self._payload_codec:
@@ -233,7 +225,7 @@ class BaseSession(ObservableMixin):
         if msg.enc_algo:
 
             if not self._payload_codec:
-                log_msg = u"received encoded payload, but no payload codec active"
+                log_msg = "received encoded payload, but no payload codec active"
                 self.log.warn(log_msg)
                 enc_err = ApplicationError(ApplicationError.ENC_NO_PAYLOAD_CODEC, log_msg, enc_algo=msg.enc_algo)
             else:
@@ -244,19 +236,19 @@ class BaseSession(ObservableMixin):
                     self.log.warn("failed to decrypt application payload 1: {err}", err=e)
                     enc_err = ApplicationError(
                         ApplicationError.ENC_DECRYPT_ERROR,
-                        u"failed to decrypt application payload 1: {}".format(e),
+                        "failed to decrypt application payload 1: {}".format(e),
                         enc_algo=msg.enc_algo,
                     )
                 else:
                     if msg.error != decrypted_error:
                         self.log.warn(
-                            u"URI within encrypted payload ('{decrypted_error}') does not match the envelope ('{error}')",
+                            "URI within encrypted payload ('{decrypted_error}') does not match the envelope ('{error}')",
                             decrypted_error=decrypted_error,
                             error=msg.error,
                         )
                         enc_err = ApplicationError(
                             ApplicationError.ENC_TRUSTED_URI_MISMATCH,
-                            u"URI within encrypted payload ('{}') does not match the envelope ('{}')".format(decrypted_error, msg.error),
+                            "URI within encrypted payload ('{}') does not match the envelope ('{}')".format(decrypted_error, msg.error),
                             enc_algo=msg.enc_algo,
                         )
 
@@ -327,7 +319,7 @@ class ApplicationSession(BaseSession):
         Implements :func:`autobahn.wamp.interfaces.ISession`
         """
         BaseSession.__init__(self)
-        self.config = config or types.ComponentConfig(realm=u"realm1")
+        self.config = config or types.ComponentConfig(realm="realm1")
 
         # set client role features supported and announced
         self._session_roles = role.DEFAULT_CLIENT_ROLES
@@ -408,13 +400,13 @@ class ApplicationSession(BaseSession):
         """
         Implements :func:`autobahn.wamp.interfaces.ISession.join`
         """
-        assert(realm is None or type(realm) == six.text_type)
+        assert(realm is None or type(realm) == str)
         assert(authmethods is None or type(authmethods) == list)
         if type(authmethods) == list:
             for authmethod in authmethods:
-                assert(type(authmethod) == six.text_type)
-        assert(authid is None or type(authid) == six.text_type)
-        assert(authrole is None or type(authrole) == six.text_type)
+                assert(type(authmethod) == str)
+        assert(authid is None or type(authid) == str)
+        assert(authrole is None or type(authrole) == str)
         assert(authextra is None or type(authextra) == dict)
 
         if self._session_id:
@@ -466,7 +458,7 @@ class ApplicationSession(BaseSession):
         """
         Implements :func:`autobahn.wamp.interfaces.ISession.onUserError`
         """
-        if isinstance(fail.value, exception.ApplicationError):
+        if hasattr(fail, 'value') and isinstance(fail.value, exception.ApplicationError):
             self.log.warn('{klass}.onUserError(): "{msg}"',
                           klass=self.__class__.__name__,
                           msg=fail.value.error_message())
@@ -499,6 +491,34 @@ class ApplicationSession(BaseSession):
             )
         return None
 
+    def type_check(self, func):
+        """
+        Does parameter type checking and validation against type hints
+        and appropriately tells the user code and the caller (through router).
+        """
+        async def _type_check(*args, **kwargs):
+            # Converge both args and kwargs into a dictionary
+            arguments = inspect.getcallargs(func, *args, **kwargs)
+            response = []
+            for name, kind in func.__annotations__.items():
+                if name in arguments:
+                    # if a "type" has __origin__ attribute it is a
+                    # parameterized generic
+                    if getattr(kind, "__origin__", None) == Union:
+                        expected_types = [arg.__name__ for arg in kind.__args__]
+                        if not isinstance(arguments[name], kind.__args__):
+                            response.append(
+                                "'{}' expected types={} got={}".format(name, expected_types,
+                                                                       type(arguments[name]).__name__))
+                    elif not isinstance(arguments[name], kind):
+                        response.append(
+                            "'{}' expected type={} got={}".format(name, kind.__name__, type(arguments[name]).__name__))
+            if response:
+                raise TypeCheckError(', '.join(response))
+            return await txaio.as_future(func, *args, **kwargs)
+
+        return _type_check
+
     def onMessage(self, msg):
         """
         Implements :func:`autobahn.wamp.interfaces.ITransportHandler.onMessage`
@@ -521,7 +541,7 @@ class ApplicationSession(BaseSession):
                     if res is not None:
                         self.log.info("Session denied by onWelcome")
                         reply = message.Abort(
-                            u"wamp.error.cannot_authenticate", u"{0}".format(res)
+                            "wamp.error.cannot_authenticate", "{0}".format(res)
                         )
                         self._transport.send(reply)
                         return
@@ -544,6 +564,7 @@ class ApplicationSession(BaseSession):
                         authprovider=self._authprovider,
                         authextra=msg.authextra,
                         serializer=self._transport._serializer.SERIALIZER_ID,
+                        transport=self._transport.peer_transport if hasattr(self._transport, 'peer_transport') else None,
                         resumed=msg.resumed,
                         resumable=msg.resumable,
                         resume_token=msg.resume_token,
@@ -585,7 +606,7 @@ class ApplicationSession(BaseSession):
 
                 def error(e):
                     reply = message.Abort(
-                        u"wamp.error.cannot_authenticate", u"Error calling onWelcome handler"
+                        "wamp.error.cannot_authenticate", "Error calling onWelcome handler"
                     )
                     self._transport.send(reply)
                     return self._swallow_error(e, "While firing onWelcome")
@@ -620,16 +641,16 @@ class ApplicationSession(BaseSession):
                 def success(signature):
                     if signature is None:
                         raise Exception('onChallenge user callback did not return a signature')
-                    if type(signature) == six.binary_type:
+                    if type(signature) == bytes:
                         signature = signature.decode('utf8')
-                    if type(signature) != six.text_type:
+                    if type(signature) != str:
                         raise Exception('signature must be unicode (was {})'.format(type(signature)))
                     reply = message.Authenticate(signature)
                     self._transport.send(reply)
 
                 def error(err):
                     self.onUserError(err, "Authentication failed")
-                    reply = message.Abort(u"wamp.error.cannot_authenticate", u"{0}".format(err.value))
+                    reply = message.Abort("wamp.error.cannot_authenticate", "{0}".format(err.value))
                     self._transport.send(reply)
                     # fire callback and close the transport
                     details = types.CloseDetails(reply.reason, reply.message)
@@ -798,7 +819,7 @@ class ApplicationSession(BaseSession):
                     if msg.enc_algo:
 
                         if not self._payload_codec:
-                            log_msg = u"received encoded payload, but no payload codec active"
+                            log_msg = "received encoded payload, but no payload codec active"
                             self.log.warn(log_msg)
                             enc_err = ApplicationError(ApplicationError.ENC_NO_PAYLOAD_CODEC, log_msg)
                         else:
@@ -812,7 +833,7 @@ class ApplicationSession(BaseSession):
                                 )
                                 enc_err = ApplicationError(
                                     ApplicationError.ENC_DECRYPT_ERROR,
-                                    u"failed to decrypt application payload 1: {}".format(e),
+                                    "failed to decrypt application payload 1: {}".format(e),
                                 )
                             else:
                                 if proc != decrypted_proc:
@@ -823,7 +844,7 @@ class ApplicationSession(BaseSession):
                                     )
                                     enc_err = ApplicationError(
                                         ApplicationError.ENC_TRUSTED_URI_MISMATCH,
-                                        u"URI within encrypted payload ('{}') does not match the envelope ('{}')".format(decrypted_proc, proc),
+                                        "URI within encrypted payload ('{}') does not match the envelope ('{}')".format(decrypted_proc, proc),
                                     )
 
                     if msg.progress:
@@ -915,7 +936,7 @@ class ApplicationSession(BaseSession):
 
                         if msg.enc_algo:
                             if not self._payload_codec:
-                                log_msg = u"received encrypted INVOCATION payload, but no keyring active"
+                                log_msg = "received encrypted INVOCATION payload, but no keyring active"
                                 self.log.warn(log_msg)
                                 enc_err = ApplicationError(ApplicationError.ENC_NO_PAYLOAD_CODEC, log_msg)
                             else:
@@ -941,7 +962,7 @@ class ApplicationSession(BaseSession):
                                         )
                                         enc_err = ApplicationError(
                                             ApplicationError.ENC_TRUSTED_URI_MISMATCH,
-                                            u"URI within encrypted INVOCATION payload ('{}') does not match the envelope ('{}')".format(decrypted_proc, proc),
+                                            "URI within encrypted INVOCATION payload ('{}') does not match the envelope ('{}')".format(decrypted_proc, proc),
                                         )
 
                         if enc_err:
@@ -970,16 +991,10 @@ class ApplicationSession(BaseSession):
                                         assert(args is None or type(args) in (list, tuple))
                                         assert(kwargs is None or type(kwargs) == dict)
 
-                                        if kwargs and six.PY2:
-                                            kwargs = {
-                                                k.decode('utf8'): v
-                                                for k, v in kwargs.iteritems()
-                                            }
-
                                         encoded_payload = None
                                         if msg.enc_algo:
                                             if not self._payload_codec:
-                                                raise Exception(u"trying to send encrypted payload, but no keyring active")
+                                                raise Exception("trying to send encrypted payload, but no keyring active")
                                             encoded_payload = self._payload_codec.encode(False, proc, args, kwargs)
 
                                         if encoded_payload:
@@ -1015,7 +1030,7 @@ class ApplicationSession(BaseSession):
                                 encoded_payload = None
                                 if msg.enc_algo:
                                     if not self._payload_codec:
-                                        log_msg = u"trying to send encrypted payload, but no keyring active"
+                                        log_msg = "trying to send encrypted payload, but no keyring active"
                                         self.log.warn(log_msg)
                                     else:
                                         try:
@@ -1067,15 +1082,21 @@ class ApplicationSession(BaseSession):
                                     self._transport.send(reply)
                                 except SerializationError as e:
                                     # the application-level payload returned from the invoked procedure can't be serialized
-                                    reply = message.Error(message.Invocation.MESSAGE_TYPE, msg.request, ApplicationError.INVALID_PAYLOAD,
-                                                          args=[u'success return value from invoked procedure "{0}" could not be serialized: {1}'.format(registration.procedure, e)])
-                                    self._transport.send(reply)
+                                    error_reply = message.Error(message.Invocation.MESSAGE_TYPE, msg.request, ApplicationError.INVALID_PAYLOAD,
+                                                                args=['success return value (args={}, kwargs={}) from invoked procedure "{}" could not be serialized: {}'.format(reply.args,
+                                                                                                                                                                                 reply.kwargs,
+                                                                                                                                                                                 registration.procedure,
+                                                                                                                                                                                 e)])
+                                    self._transport.send(error_reply)
                                 except PayloadExceededError as e:
                                     # the application-level payload returned from the invoked procedure, when serialized and framed
                                     # for the transport, exceeds the transport message/frame size limit
-                                    reply = message.Error(message.Invocation.MESSAGE_TYPE, msg.request, ApplicationError.PAYLOAD_SIZE_EXCEEDED,
-                                                          args=[u'success return value from invoked procedure "{0}" exceeds transport size limit: {1}'.format(registration.procedure, e)])
-                                    self._transport.send(reply)
+                                    error_reply = message.Error(message.Invocation.MESSAGE_TYPE, msg.request, ApplicationError.PAYLOAD_SIZE_EXCEEDED,
+                                                                args=['success return value (args={}, kwargs={}) from invoked procedure "{}" exceeds transport size limit: {}'.format(reply.args,
+                                                                                                                                                                                      reply.kwargs,
+                                                                                                                                                                                      registration.procedure,
+                                                                                                                                                                                      e)])
+                                    self._transport.send(error_reply)
 
                             def error(err):
                                 del self._invocations[msg.request]
@@ -1104,13 +1125,13 @@ class ApplicationSession(BaseSession):
                                 except SerializationError as e:
                                     # the application-level payload returned from the invoked procedure can't be serialized
                                     reply = message.Error(message.Invocation.MESSAGE_TYPE, msg.request, ApplicationError.INVALID_PAYLOAD,
-                                                          args=[u'error return value from invoked procedure "{0}" could not be serialized: {1}'.format(registration.procedure, e)])
+                                                          args=['error return value from invoked procedure "{0}" could not be serialized: {1}'.format(registration.procedure, e)])
                                     self._transport.send(reply)
                                 except PayloadExceededError as e:
                                     # the application-level payload returned from the invoked procedure, when serialized and framed
                                     # for the transport, exceeds the transport message/frame size limit
                                     reply = message.Error(message.Invocation.MESSAGE_TYPE, msg.request, ApplicationError.PAYLOAD_SIZE_EXCEEDED,
-                                                          args=[u'success return value from invoked procedure "{0}" exceeds transport size limit: {1}'.format(registration.procedure, e)])
+                                                          args=['success return value from invoked procedure "{0}" exceeds transport size limit: {1}'.format(registration.procedure, e)])
                                     self._transport.send(reply)
 
                                 # we have handled the error, so we eat it
@@ -1164,7 +1185,7 @@ class ApplicationSession(BaseSession):
                             " ID {0}".format(msg.registration)
                         )
                     self.log.info(
-                        u"Router unregistered procedure '{proc}' with ID {id}",
+                        "Router unregistered procedure '{proc}' with ID {id}",
                         proc=reg.procedure,
                         id=msg.registration,
                     )
@@ -1233,7 +1254,7 @@ class ApplicationSession(BaseSession):
             # fire callback and close the transport
             details = types.CloseDetails(
                 reason=types.CloseDetails.REASON_TRANSPORT_LOST,
-                message=u'WAMP transport was lost without closing the session {} before'.format(self._session_id),
+                message='WAMP transport was lost without closing the session {} before'.format(self._session_id),
             )
             d = txaio.as_future(self.onLeave, details)
 
@@ -1339,11 +1360,11 @@ class ApplicationSession(BaseSession):
         Implements :func:`autobahn.wamp.interfaces.ISession.leave`
         """
         if not self._session_id:
-            raise SessionNotReady(u"session hasn't joined a realm")
+            raise SessionNotReady("session hasn't joined a realm")
 
         if not self._goodbye_sent:
             if not reason:
-                reason = u"wamp.close.normal"
+                reason = "wamp.close.normal"
             msg = wamp.message.Goodbye(reason=reason, message=message)
             self._transport.send(msg)
             self._goodbye_sent = True
@@ -1370,7 +1391,7 @@ class ApplicationSession(BaseSession):
         """
         Implements :func:`autobahn.wamp.interfaces.IPublisher.publish`
         """
-        assert(type(topic) == six.text_type)
+        assert(type(topic) == str)
         assert(args is None or type(args) in (list, tuple))
         assert(kwargs is None or type(kwargs) == dict)
 
@@ -1384,12 +1405,6 @@ class ApplicationSession(BaseSession):
         if options and not isinstance(options, types.PublishOptions):
             raise Exception("options must be of type a.w.t.PublishOptions")
 
-        if kwargs and six.PY2:
-            kwargs = {
-                k.decode('utf8'): v
-                for k, v in kwargs.iteritems()
-            }
-
         if not self._transport:
             raise exception.TransportLost()
 
@@ -1463,18 +1478,18 @@ class ApplicationSession(BaseSession):
         return on_reply
 
     @public
-    def subscribe(self, handler, topic=None, options=None):
+    def subscribe(self, handler, topic=None, options=None, check_types=False):
         """
         Implements :func:`autobahn.wamp.interfaces.ISubscriber.subscribe`
         """
-        assert((callable(handler) and topic is not None) or hasattr(handler, '__class__'))
-        assert(topic is None or type(topic) == six.text_type)
+        assert((callable(handler) and topic is not None) or (hasattr(handler, '__class__') and not check_types))
+        assert(topic is None or type(topic) == str)
         assert(options is None or isinstance(options, types.SubscribeOptions))
 
         if not self._transport:
             raise exception.TransportLost()
 
-        def _subscribe(obj, fn, topic, options):
+        def _subscribe(obj, fn, topic, options, check_types):
             message.check_or_raise_uri(topic,
                                        message='{}.subscribe()'.format(self.__class__.__name__),
                                        strict=False,
@@ -1483,6 +1498,8 @@ class ApplicationSession(BaseSession):
 
             request_id = self._request_id_gen.next()
             on_reply = txaio.create_future()
+            if check_types:
+                fn = self.type_check(fn)
             handler_obj = Handler(fn, obj, options.details_arg if options else None)
             self._subscribe_reqs[request_id] = SubscribeRequest(request_id, topic, on_reply, handler_obj)
 
@@ -1506,7 +1523,7 @@ class ApplicationSession(BaseSession):
 
         if callable(handler):
             # subscribe a single handler
-            return _subscribe(None, handler, topic, options)
+            return _subscribe(None, handler, topic, options, check_types)
 
         else:
 
@@ -1521,10 +1538,10 @@ class ApplicationSession(BaseSession):
                             subopts = pat.options or options
                             if subopts is None:
                                 if pat.uri_type == uri.Pattern.URI_TYPE_WILDCARD:
-                                    subopts = types.SubscribeOptions(match=u"wildcard")
+                                    subopts = types.SubscribeOptions(match="wildcard")
                                 else:
-                                    subopts = types.SubscribeOptions(match=u"exact")
-                            on_replies.append(_subscribe(handler, proc, _uri, subopts))
+                                    subopts = types.SubscribeOptions(match="exact")
+                            on_replies.append(_subscribe(handler, proc, _uri, subopts, pat._check_types))
 
             # XXX needs coverage
             return txaio.gather(on_replies, consume_exceptions=True)
@@ -1568,7 +1585,7 @@ class ApplicationSession(BaseSession):
         """
         Implements :func:`autobahn.wamp.interfaces.ICaller.call`
         """
-        assert(type(procedure) == six.text_type)
+        assert(type(procedure) == str)
         assert(args is None or type(args) in (list, tuple))
         assert(kwargs is None or type(kwargs) == dict)
 
@@ -1582,12 +1599,6 @@ class ApplicationSession(BaseSession):
         if options and not isinstance(options, types.CallOptions):
             raise Exception("options must be of type a.w.t.CallOptions")
 
-        if kwargs and six.PY2:
-            kwargs = {
-                k.decode('utf8'): v
-                for k, v in kwargs.iteritems()
-            }
-
         if not self._transport:
             raise exception.TransportLost()
 
@@ -1668,19 +1679,19 @@ class ApplicationSession(BaseSession):
         return on_reply
 
     @public
-    def register(self, endpoint, procedure=None, options=None, prefix=None):
+    def register(self, endpoint, procedure=None, options=None, prefix=None, check_types=False):
         """
         Implements :func:`autobahn.wamp.interfaces.ICallee.register`
         """
-        assert((callable(endpoint) and procedure is not None) or hasattr(endpoint, '__class__'))
-        assert(procedure is None or type(procedure) == six.text_type)
+        assert((callable(endpoint) and procedure is not None) or (hasattr(endpoint, '__class__') and not check_types))
+        assert(procedure is None or type(procedure) == str)
         assert(options is None or isinstance(options, types.RegisterOptions))
-        assert prefix is None or isinstance(prefix, six.text_type)
+        assert prefix is None or isinstance(prefix, str)
 
         if not self._transport:
             raise exception.TransportLost()
 
-        def _register(obj, fn, procedure, options):
+        def _register(obj, fn, procedure, options, check_types):
             message.check_or_raise_uri(procedure,
                                        message='{}.register()'.format(self.__class__.__name__),
                                        strict=False,
@@ -1689,9 +1700,11 @@ class ApplicationSession(BaseSession):
 
             request_id = self._request_id_gen.next()
             on_reply = txaio.create_future()
+            if check_types:
+                fn = self.type_check(fn)
             endpoint_obj = Endpoint(fn, obj, options.details_arg if options else None)
             if prefix is not None:
-                procedure = u"{}{}".format(prefix, procedure)
+                procedure = "{}{}".format(prefix, procedure)
             self._register_reqs[request_id] = RegisterRequest(request_id, on_reply, procedure, endpoint_obj)
 
             if options:
@@ -1715,7 +1728,7 @@ class ApplicationSession(BaseSession):
         if callable(endpoint):
 
             # register a single callable
-            return _register(None, endpoint, procedure, options)
+            return _register(None, endpoint, procedure, options, check_types)
 
         else:
 
@@ -1728,7 +1741,7 @@ class ApplicationSession(BaseSession):
                         if pat.is_endpoint():
                             _uri = pat.uri()
                             regopts = pat.options or options
-                            on_replies.append(_register(endpoint, proc, _uri, regopts))
+                            on_replies.append(_register(endpoint, proc, _uri, regopts, pat._check_types))
 
             # XXX needs coverage
             return txaio.gather(on_replies, consume_exceptions=True)
@@ -1764,7 +1777,7 @@ class _SessionShim(ApplicationSession):
     **NOTE:** this is not public or intended for use; you should import
     either :class:`autobahn.asyncio.wamp.Session` or
     :class:`autobahn.twisted.wamp.Session` depending on which async
-    framework you're using.
+    framework yo're using.
     """
 
     #: name -> IAuthenticator
@@ -1787,8 +1800,8 @@ class _SessionShim(ApplicationSession):
             self.join(
                 self.config.realm,
                 authmethods=list(self._authenticators.keys()),
-                authid=authid or u'public',
-                authrole=authrole or u'default',
+                authid=authid or 'public',
+                authrole=authrole or 'default',
                 authextra=authextra,
             )
         else:
@@ -1945,7 +1958,7 @@ class ApplicationSessionFactory(object):
         :param config: The default component configuration.
         :type config: instance of :class:`autobahn.wamp.types.ComponentConfig`
         """
-        self.config = config or types.ComponentConfig(realm=u"realm1")
+        self.config = config or types.ComponentConfig(realm="realm1")
 
     def __call__(self):
         """
